@@ -34,7 +34,8 @@ class SpectralNormLinear(nn.Module):
 
     def reset_parameters(self) -> None:
         weights = torch.empty(self.out_features, self.in_features)
-        nn.init.kaiming_uniform_(weights, a=math.sqrt(5))
+        stdv = 1.0 / math.sqrt(self.in_features)
+        nn.init.uniform_(weights, a=-stdv, b=stdv)
         if self.use_bias:
             fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weights)
             bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
@@ -77,7 +78,8 @@ class AnchoredLinear(nn.Module):
 
     def reset_parameters(self) -> None:
         weights = torch.empty(self.out_features, self.in_features)
-        nn.init.kaiming_uniform_(weights, a=math.sqrt(5))
+        stdv = 1.0 / math.sqrt(self.in_features)
+        nn.init.uniform_(weights, a=-stdv, b=stdv)
         if self.use_bias:
             fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weights)
             bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
@@ -95,7 +97,7 @@ class AnchoredLinear(nn.Module):
         )
 
 
-class ReparamLinear(nn.Module):
+class WeightNormedLinear(nn.Module):
     """
     ...
     """
@@ -120,7 +122,8 @@ class ReparamLinear(nn.Module):
 
     def reset_parameters(self) -> None:
         weights = torch.empty(self.out_features, self.in_features)
-        nn.init.kaiming_uniform_(weights, a=math.sqrt(5))
+        stdv = 1.0 / math.sqrt(self.in_features)
+        nn.init.uniform_(weights, a=-stdv, b=stdv)
         if self.use_bias:
             fan_in, _ = nn.init._calculate_fan_in_and_fan_out(weights)
             bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
@@ -131,8 +134,7 @@ class ReparamLinear(nn.Module):
         return F.linear(x, self.weights(), self.bias)
 
     def __repr__(self) -> str:
-        # Optional: A nice representation for printing the module.
         return (
-            f"AnchoredLinear(in_features={self.in_features},"
+            f"WeightNormedLinear(in_features={self.in_features},"
             f"out_features={self.out_features}, bias={self.use_bias})"
         )
