@@ -53,12 +53,20 @@ class ClassificationHead(nn.Module):
     A general classification head for a ViT
     """
 
-    def __init__(self, d_model, linear_module, n_classes, batch_norm=True):
+    def __init__(
+        self,
+        d_model,
+        n_classes,
+        linear_module=nn.Linear,
+        logit_projection_layer=nn.Linear,
+        batch_norm_logits=True,
+    ):
         super().__init__()
         self.d_model = d_model
         self.summarize = GetCLSToken()
-        self.projection = nn.Linear(d_model, n_classes)
-        if batch_norm:
+        self.projection = logit_projection_layer(d_model, n_classes)
+
+        if batch_norm_logits:
             self.batch_norm = nn.BatchNorm1d(n_classes, affine=False)
         else:
             self.batch_norm = nn.Identity()
@@ -83,8 +91,16 @@ class SequencePoolClassificationHead(ClassificationHead):
     """
 
     def __init__(self, d_model, linear_module, out_dim, batch_norm=True):
-        super().__init__(d_model, linear_module, out_dim, batch_norm=batch_norm)
+        super().__init__(d_model, linear_module, out_dim, batch_norm_logits=batch_norm)
         self.summarize = SequencePool(d_model, linear_module)
+        # Rebuild the classification process with the correct summary module:
+        self.classification_process = nn.Sequential(
+            *[
+                self.summarize,
+                self.projection,
+                self.batch_norm,
+            ]
+        )
 
 
 class ViTEncoder(nn.Module):
@@ -117,8 +133,13 @@ class ViTEncoder(nn.Module):
         pooling_kernel_stride=2,
         pooling_padding=1,
         transformer_feedforward_first=True,
+        transformer_initial_ff_residual_path=True,
+        transformer_initial_ff_linear_module_up=None,
+        transformer_initial_ff_linear_module_down=None,
+        transformer_initial_ff_mlp_dropout=None,
         transformer_pre_norm=True,
         transformer_normformer=False,
+        transformer_post_norm=False,
         transformer_position_embedding="relative",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
@@ -128,6 +149,9 @@ class ViTEncoder(nn.Module):
         transformer_return_bos_tokens=False,
         transformer_activation: nn.Module = SquaredReLU,
         transformer_activation_kwargs: Optional[dict] = None,
+        transformer_ff_linear_module_up=None,
+        transformer_ff_linear_module_down=None,
+        transformer_msa_scaling="d",
         transformer_mlp_dropout=0.0,
         transformer_msa_dropout=0.1,
         transformer_stochastic_depth=0.1,
@@ -232,13 +256,7 @@ class ViTEncoder(nn.Module):
 
         if pooling_type is None:
             pooling_out_channels = cnn_activation_out_channels
-            self.pool = nn.Sequential(
-                *[
-                    Rearrange(
-                        f"N C {spatial_dim_names} -> N ({spatial_dim_names}) C"
-                    ),  # for transformer
-                ]
-            )
+            self.pool = nn.Identity()
 
         elif pooling_type == "max":
             pooling_out_channels = cnn_activation_out_channels
@@ -284,6 +302,9 @@ class ViTEncoder(nn.Module):
                 mlp_ratio=transformer_mlp_ratio,
                 activation=transformer_activation,
                 activation_kwargs=transformer_activation_kwargs,
+                ff_linear_module_up=transformer_ff_linear_module_up,
+                ff_linear_module_down=transformer_ff_linear_module_down,
+                msa_scaling=transformer_msa_scaling,
                 mlp_dropout=transformer_mlp_dropout,
                 msa_dropout=transformer_msa_dropout,
                 stochastic_depth=transformer_stochastic_depth,
@@ -293,6 +314,7 @@ class ViTEncoder(nn.Module):
                 return_bos_tokens=transformer_return_bos_tokens,
                 pre_norm=transformer_pre_norm,
                 normformer=transformer_normformer,
+                post_norm=transformer_post_norm,
             )
         else:
             self.transformer = nn.Identity()
@@ -304,11 +326,28 @@ class ViTEncoder(nn.Module):
                 transformer_embedding_size,
                 activation=transformer_activation,
                 activation_kwargs=transformer_activation_kwargs,
-                dropout=transformer_mlp_dropout,
-                linear_module=linear_module,
+                dropout=(
+                    # First truthy assigned value
+                    transformer_initial_ff_mlp_dropout
+                    if transformer_initial_ff_mlp_dropout is not None
+                    else transformer_mlp_dropout
+                ),
+                linear_module_up=(
+                    # First truthy assigned value
+                    transformer_initial_ff_linear_module_up
+                    or transformer_ff_linear_module_up
+                    or linear_module
+                ),
+                linear_module_down=(
+                    # First truthy assigned value
+                    transformer_initial_ff_linear_module_down
+                    or transformer_ff_linear_module_down
+                    or linear_module
+                ),
                 pre_norm=transformer_pre_norm,
                 normformer=transformer_normformer,
-                raw_input=not cnn,
+                post_norm=transformer_post_norm,
+                residual_path=transformer_initial_ff_residual_path,
             )
         else:
             self.initial_ff = nn.Identity()
@@ -334,11 +373,7 @@ class ViTEncoder(nn.Module):
 
 class ViT(nn.Module):
     """
-    Denoising convolutional transformer
-    Based on the Compact Convolutional Transformer (CCT) of [Hasani et al. (2021)
-        *''Escaping the Big Data Paradigm with Compact Transformers''*](
-        https://arxiv.org/abs/2104.05704). It's a convolutional neural network
-        leading into a transformer encoder, followed by a sequence pooling layer.
+    ...
     """
 
     def __init__(
@@ -362,8 +397,13 @@ class ViT(nn.Module):
         pooling_kernel_stride=2,
         pooling_padding=1,
         transformer_feedforward_first=True,
+        transformer_initial_ff_residual_path=True,
+        transformer_initial_ff_linear_module_up=None,
+        transformer_initial_ff_linear_module_down=None,
+        transformer_initial_ff_mlp_dropout=None,
         transformer_pre_norm=True,
         transformer_normformer=False,
+        transformer_post_norm=False,
         transformer_position_embedding="relative",  # absolute or relative
         transformer_embedding_size=256,
         transformer_layers=7,
@@ -373,12 +413,16 @@ class ViT(nn.Module):
         transformer_return_bos_tokens=False,
         transformer_activation: nn.Module = SquaredReLU,
         transformer_activation_kwargs: Optional[dict] = None,
+        transformer_ff_linear_module_up=None,
+        transformer_ff_linear_module_down=None,
+        transformer_msa_scaling="d",
         transformer_mlp_dropout=0.0,
         transformer_msa_dropout=0.1,
         transformer_stochastic_depth=0.1,
-        batch_norm_outputs=True,
-        linear_module=SpectralNormLinear,
         head=SequencePoolClassificationHead,
+        batch_norm_logits=True,
+        logit_projection_layer=nn.Linear,
+        linear_module=nn.Linear,
     ):
 
         super().__init__()
@@ -418,8 +462,13 @@ class ViT(nn.Module):
             pooling_kernel_stride=pooling_kernel_stride,
             pooling_padding=pooling_padding,
             transformer_feedforward_first=transformer_feedforward_first,
+            transformer_initial_ff_residual_path=transformer_initial_ff_residual_path,
+            transformer_initial_ff_linear_module_up=transformer_initial_ff_linear_module_up,
+            transformer_initial_ff_linear_module_down=transformer_initial_ff_linear_module_down,
+            transformer_initial_ff_mlp_dropout=transformer_initial_ff_mlp_dropout,
             transformer_pre_norm=transformer_pre_norm,
             transformer_normformer=transformer_normformer,
+            transformer_post_norm=transformer_post_norm,
             transformer_position_embedding=transformer_position_embedding,
             transformer_embedding_size=transformer_embedding_size,
             transformer_layers=transformer_layers,
@@ -429,6 +478,9 @@ class ViT(nn.Module):
             transformer_return_bos_tokens=transformer_return_bos_tokens,
             transformer_activation=transformer_activation,
             transformer_activation_kwargs=transformer_activation_kwargs,
+            transformer_ff_linear_module_up=transformer_ff_linear_module_up,
+            transformer_ff_linear_module_down=transformer_ff_linear_module_down,
+            transformer_msa_scaling=transformer_msa_scaling,
             transformer_mlp_dropout=transformer_mlp_dropout,
             transformer_msa_dropout=transformer_msa_dropout,
             transformer_stochastic_depth=transformer_stochastic_depth,
@@ -437,9 +489,10 @@ class ViT(nn.Module):
 
         self.pool = head(
             transformer_embedding_size,
-            linear_module,
             image_classes,
-            batch_norm=batch_norm_outputs,
+            linear_module=linear_module,
+            logit_projection_layer=logit_projection_layer,
+            batch_norm=batch_norm_logits,
         )
 
     @property
